from io import BytesIO

import plone.api
from AccessControl.SecurityInfo import ClassSecurityInfo
from DateTime import DateTime
from OFS.Folder import Folder
from Products.Five import BrowserView
from plone.protect.interfaces import IDisableCSRFProtection
from plone.rfc822.interfaces import IPrimaryFieldInfo
from zope.interface import implementer
from zope.interface.declarations import alsoProvides

from .interfaces import ITrashCan
from .permissions import ManageTrash
from .trash import TrashedItem

try:
    from ims.upload.content import Chunk, ChunkedFile
except NameError:
    class Chunk(object):
        pass
MAX_TRASH_SIZE = 1e9


def generate_id(start_id):
    now = DateTime()
    time_str = '%s.%s' % (now.strftime('%Y-%m-%d'), str(now.millis())[7:])
    return start_id + '-' + time_str


@implementer(ITrashCan)
class PloneTrashCan(Folder):
    """Plone Trash Can"""
    security = ClassSecurityInfo()
    disposal_frequency = 7

    manage_options = [{'label': 'Trash Can', 'action': 'manage_main'},
                      {'label': 'Configuration', 'action': 'manage_propertiesForm'}]

    _properties = ({'id': 'disposal_frequency', 'type': 'int', 'mode': 'w'},)

    @staticmethod
    def trashable(ob):
        """ Don't trash anything too large. Only works on primary field objects """
        if isinstance(ob, Chunk):
            return False
        elif isinstance(ob, ChunkedFile):
            return False

        try:
            primary_field = IPrimaryFieldInfo(ob)
        except TypeError:
            return True
        else:
            if primary_field and \
                    primary_field.value and \
                    hasattr(primary_field.value, 'size') and \
                    primary_field.value.size > MAX_TRASH_SIZE:
                return False
            return True

    def trash(self, ob):
        """ trash the item """
        if self.trashable(ob):
            id = generate_id(ob.getId())
            title = ob.Title()
            data = self.zexpickle(ob)
            opath = '/'.join(ob.getPhysicalPath()[:-1])
            _trash = TrashedItem(id, title, data, opath)
            self._setObject(id, _trash, suppress_events=True)

    def restore(self, id):
        """ restore the item to its original path
            if it exists, otherwise root """
        _trash = self[id]
        data = _trash.data.open("r").read()
        opath = _trash.path
        try:
            source = self.restrictedTraverse(opath)
        except KeyError:  # Path is invalid, likely the container was moved or deleted. Restore to root
            source = plone.api.portal.get()
        ob = self.unzexpickle(data)
        source._setObject(id[:-18], ob)
        self._delObject(id)

    def zexpickle(self, ob):
        """ pickle + add zexp wrapper """
        f = BytesIO()
        ob._p_jar.exportFile(ob._p_oid, f)
        data = f.getvalue()
        return data

    def unzexpickle(self, data):
        """ unpickle + remove zexp wrapper """
        f = BytesIO()
        f.write(data)

        # we HAVE to skip the first four characters, which are just
        # 'ZEXP', because _importDuringCommit does length validation
        # with the assumption that the file stream is already at this
        # point, not at SOF
        f.seek(4)

        conn = self._p_jar
        import transaction as t

        return_oid_list = []
        conn._import = f, return_oid_list
        conn._register()

        t.savepoint(optimistic=True)
        if return_oid_list:
            return conn.get(return_oid_list[0])
        else:
            return None

    @security.protected(ManageTrash)
    def deleteExpired(self):
        """Delete all of the content that is expired
           Content expires when it is older than X days, where X is the disposal_frequency property
        """
        expiredDate = DateTime() - self.disposal_frequency
        for trash in self.objectValues():
            # "or not trash.created()" is to handle legacy items
            if expiredDate > trash.created() or not trash.created():
                self._delObject(trash.getId())

    @security.protected(ManageTrash)
    def manage_restore(self, id, REQUEST=None):
        """Attempts to copy the trashed item to its original path"""
        self.restore(id)

        msg = '%s has been restored.' % id
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, manage_tabs_message=msg)

    __call__ = deleteExpired


class TrashCanClear(BrowserView):
    def __call__(self):
        alsoProvides(self.request, IDisableCSRFProtection)
        self.context.deleteExpired()
        return "content deleted"



from plone.app.testing import PloneSandboxLayer, IntegrationTesting, FunctionalTesting, applyProfile
from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
import ims.trashcan
import ims.upload
import collective.js.jqueryui

class TrashcanSiteLayer(PloneSandboxLayer):
    defaultBases = (PLONE_APP_CONTENTTYPES_FIXTURE,)

    def setUpZope(self, app, configuration_context):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base
        # layer.
        self.loadZCML(package=collective.js.jqueryui)
        self.loadZCML(package=ims.upload)
        self.loadZCML(package=ims.trashcan)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ims.trashcan:default')
        applyProfile(portal, 'ims.upload:default')


TRASHCAN_SITE_FIXTURE = TrashcanSiteLayer()

INTEGRATION = IntegrationTesting(
    bases=(TRASHCAN_SITE_FIXTURE,),
    name="ims.trashcan:Integration"
)

FUNCTIONAL = FunctionalTesting(
    bases=(TRASHCAN_SITE_FIXTURE,),
    name="ims.trashcan:Functional"
)



import plone.api
from plone.app.textfield.value import RichTextValue

from plone.namedfile.file import NamedBlobFile
from . import base


class TestTrashCan(base.IntegrationTestCase):

    def test_document(self):
        page = plone.api.content.create(id='page1', type='Document', title='My page', container=self.portal)
        page.text = RichTextValue('some text')
        plone.api.content.delete(page)
        self.assertEqual(len(self.can.objectIds()), 1)

    def test_file(self):
        _file = plone.api.content.create(id='file1', type='File', title='My file', container=self.portal)
        plone.api.content.delete(_file)
        self.assertEqual(len(self.can.objectIds()), 1)

    def test_chunk(self):
        cfile = plone.api.content.create(id='cf1', type='ChunkedFile', title='My file', container=self.portal)
        chunk = plone.api.content.create(id='chunk1', type='Chunk', title='chonky boi', container=cfile)
        plone.api.content.delete(chunk)
        self.assertEqual(len(self.can.objectIds()), 0)
        plone.api.content.delete(cfile)
        self.assertEqual(len(self.can.objectIds()), 0)

    def test_file_limit(self):
        _file = plone.api.content.create(id='file1', type='File', title='My file', container=self.portal)
        _file.file = NamedBlobFile()
        _file.file.data = 'hello, world'
        _file.file.__dict__['size'] = 1e10  # fake it
        plone.api.content.delete(_file)
        self.assertEqual(len(self.can.objectIds()), 0)


def test_suite():
    import unittest
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



import plone.api


def trashEventHandler(ob, event):
    """ Item is deleted - trash it!"""
    try:
        can = plone.api.portal.get_tool('portal_trash_can')
    except plone.api.exc.InvalidParameterError:
        pass  # not installed
    except plone.api.exc.CannotGetPortalError:
        pass  # portal is being deleted
    else:
        can.trash(ob)


def trashAdded(ob, event):
    """ nothing for now """
    pass


def trashRemoved(ob, event):
    """ nothing for now """
    pass



import plone.api
from DateTime import DateTime
from OFS.ObjectManager import ObjectManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from ZODB.blob import Blob
from zope.interface.declarations import implementer

from .interfaces import ITrashedItem


@implementer(ITrashedItem)
class TrashedItem(ObjectManager):
    manage_main = PageTemplateFile('www/manage_trashedItem', globals())
    _created = ''

    def __init__(self, id, title='', data=None, path=''):
        self.id = id
        self.data = Blob(data)
        self.path = path
        self.Title = title
        self._created = DateTime()

    def manage_properties(self):
        return {
            'title': self.Title,
            'created': plone.api.portal.get_localized_time(self._created, long_format=True),
            'id': self.id,
            'path': self.path,
        }

    def created(self):
        return self._created

    def getId(self):
        """ needed for indexing """
        return self.id



from zope.interface.interface import Interface
from zope.schema import Bytes, TextLine


class ITrashedItem(Interface):
    """ deleted item with trash wrapper """
    id = TextLine(
        title="ID",
        description="ID",
    )

    Title = TextLine(
        title="Title",
        description="Original title",
    )

    data = Bytes(
        title="Data",
        description="The fake zexp file",
    )

    path = TextLine(
        title="Path",
        description="The path to its restore point",
    )


class ITrashCan(Interface):
    """ marker """



from Products.CMFPlone.utils import get_installer
from plone import api
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles

from . import base
from .. import testing


class TestUninstall(base.IntegrationTestCase):
    layer = testing.INTEGRATION

    def setUp(self):
        self.portal = self.layer['portal']
        roles_before = api.user.get_roles(TEST_USER_ID)
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        self.installer = get_installer(self.portal)
        self.installer.uninstall_product('ims.tools')
        setRoles(self.portal, TEST_USER_ID, roles_before)

    def test_product_uninstalled(self):
        """Test if ims.tools is cleanly uninstalled."""
        self.assertFalse(self.installer.is_product_installed('ims.tools'))



import unittest

from ims.trashcan import testing
from plone.app.testing import setRoles, TEST_USER_ID, SITE_OWNER_NAME, SITE_OWNER_PASSWORD
from plone.testing.zope import Browser


class UnitTestCase(unittest.TestCase):
    def setUp(self):
        pass


class IntegrationTestCase(unittest.TestCase):
    layer = testing.INTEGRATION

    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        self.portal = self.layer['portal']
        self.can = self.portal.portal_trash_can
        self.request = self.layer
        setRoles(self.portal, TEST_USER_ID, ['Manager'])


class FunctionalTestCase(IntegrationTestCase):
    layer = testing.FUNCTIONAL

    def setUp(self):
        super(FunctionalTestCase, self).setUp()
        self.browser = Browser(self.layer['app'])
        self.browser.handleErrors = False
        self.browser.addHeader(
            'Authorization',
            'Basic %s:%s' % (SITE_OWNER_NAME, SITE_OWNER_PASSWORD,)
        )



from setuptools import setup, find_packages

version = '4.1.2'

setup(name='ims.trashcan',
      version=version,
      description="Stores deleted objects in a trash can before deletion.",
      long_description=open("README.md").read(),
      long_description_content_type='text/markdown',
      classifiers=[
          "Framework :: Plone :: 5.2",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
      ],
      keywords='',
      author='Eric Wohnlich',
      author_email='wohnlice@imsweb.com',
      url='https://git.imsweb.com/plone/ims.trashcan',
      license='MIT',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ims'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
      ],
      entry_points="""
      # -*- Entry points: -*-

      [z3c.autoinclude.plugin]
      target = plone
      """,
      extras_require={
          'test': ['plone.app.testing', 'plone.mocktestcase', 'formencode'],
      },
      )



from Products.CMFCore import utils

from . import can


def initialize(context):
    utils.ToolInit(
        'ims.trashcan tool',
        tools=(can.PloneTrashCan,),
        icon='tool.png', ).initialize(context)
